%% Code accompanying "Uncertainty Quantification and Global Sensitivity Analysis for Economic Models"
%% by Daniel Harenberg, Stefano Marelli, Bruno Sudret, and Viktor Winschel

%% Main file, final version for publication
% Last date modified: 29th March 2018

%% delete figures in result folder
if (exist('results','dir'))
    delete('results/*.*');
else
    mkdir('results')
end
diary('results/diary.txt');
diary on
tic
%% Sensitivity analysis for the output of the economic growth model
clearvars;          % clear workspace
uqlab;              % start uqlab
rng(100);           % reseed random number generator
clc;                % clear command window

UQSESSION = 'UQLab_session.mat';
%% return residual values from compecon 
global resids

global additional_QoIs % compute more than just the first two QoIs.
additional_QoIs = false;

%% 1 - Specify the probabilistic input
parNames{1} = '\beta';  % Discount factor                         
parNames{2} = '\tau';   % Elasticity of intertemporal substitution
parNames{3} = '\chi'; % Consumption and leisure substitution rate
parNames{4} = '\alpha'; % Technical substitution rate              
parNames{5} = '\delta'; % Depreciation rate                        
parNames{6} = '\phi';   % Capital adjustment cost parameter        
parNames{7} = '\rho';   % Autocorrelation of productivity          
parNames{8} = '\sigma'; % Standard deviation of productivity shock 

% parameter bounds          baseline values
parVals = [0.97,   0.99;        % beta  0.98
           1.00,   3.00;        % tau   2
           0.2,    0.4;         % theta 0.3
           0.2,    0.5;         % alpha 0.35
           0.01,   0.03;        % delta 0.02
           0.00,   1.0;         % phi   0.5
           0.92,   0.98;        % rho   0.95
           0.005,  0.009];      % sigma 0.007

    
for i=1:8                       % Independent marginals
    parNameTex = parNames{i};
    Input.Marginals(i).Name = parNameTex(2:end);
    Input.Marginals(i).Type = 'Uniform';
    Input.Marginals(i).Parameters = parVals(i,:);
end
myInput = uq_createInput(Input);    % Create input object

% model outputs
outNames = {'Average capital-output ratio', 'Production variance ratio'}; 
if (additional_QoIs)
    outNames = {'Average capital-output ratio', ...
                'Production variance ratio', ...
                'Consumption variance ratio', ...
                'Corr(cons,production)', ...
                'Variance consumption growth',... %}; 
                'Corr(investment,production)', ...
                'Autocorr(investment)'};
end    

% some reused parameter vectors
parC=[0.98,2,0.3,0.35,0.02,0.5,0.95,0.007];  % baseline values
parU=parVals(:,2)';                           % upper bounds
parL=parVals(:,1)';                           % lower bounds

%% Scenario Analysis
%
parS1=parC; parS2=parC; parS3=parC;           % preallocate with baseline parmaeters
parS1([2,7,8]) = parU([2,7,8]);               % Scenario 1: high risk, high risk aversion
parS2([4,5,6]) = parU([4,5,6]);               % Scenario 2: high capital utilization, high frictions
%parS3([1,2,3]) = parL([1,2,3]);               % Scenario 3: enjoy life preferences
%sY = M_solve_RBC([parC;parS1;parS2;parS3]);
sY = M_solve_RBC([parC;parS1;parS2]);   
plot(sY(:,2),sY(:,1),'bd','MarkerSize',8,'MarkerFaceColor','b');
fontsize = 20;
set(gca,'fontsize',fontsize-4)
%title('Scenario Analysis');
xlabel(outNames{2},'FontSize',fontsize);
ylabel(outNames{1},'FontSize',fontsize);
text(sY(1,2),sY(1,1),' Baseline','FontSize',fontsize);
text(sY(2,2),sY(2,1),' High Risk','FontSize',fontsize);
text(sY(3,2),sY(3,1),' High Utilization','FontSize',fontsize);
%text(sY(4,2),sY(4,1),' Enjoy Life','FontSize',fontsize);
xlim([0.5*min(sY(:,2)) 1.5*max(sY(:,2))]); % 0.0
ylim([0.8*min(sY(:,1)) 1.2*max(sY(:,1))]); %0.0
saveas(gcf, 'results/sY_Scenario_analysis.fig');
saveas(gcf, 'results/sY_Scenario_analysis.png');
close(gcf);

%% Local Sensitivity Analysis
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
disp('---------------------------------------')
disp('LOCAL SENSITIVITY ANALYSIS')
disp('---------------------------------------')
disp('')
%% one at a time finite differences
disp('---------------------------------------')
disp('Calculate one at a time percentage changes...')
%
disp('mean and variance: one at a time baseline to upper parameters values:');
dY = M_oaat_findiff(parC,parU);
M_plot_sens(dY,outNames,parNames,'One-at-a-Time Percentage Changes','results/sY_M_V_Sens_oaat_can_up');
disp(sprintf('%s    & ',parNames{:}));
disp(sprintf('%3.3f & ',dY(:,1)));
disp(sprintf('%3.3f & ',dY(:,2)));
left_panel = dY;

disp('mean and variance: one at a time lower to baseline parameters values:');
dY = M_oaat_findiff(parL,parC);
M_plot_sens(dY,outNames,parNames,'One-at-a-Time Percentage Changes','results/sY_M_V_Sens_oaat_low_can');
disp(sprintf('%s    & ',parNames{:}));
disp(sprintf('%3.3f & ',dY(:,1)));
disp(sprintf('%3.3f & ',dY(:,2)));
right_panel=dY;

save('results/fig2_values.mat', 'left_panel', 'right_panel');
plot_OATs_sidebyside

%% Global Sensitivity Analysis
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
disp('---------------------------------------')
disp('GLOBAL SENSITIVITY ANALYSIS')
disp('---------------------------------------')
disp('')

%% %%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%
% stochastic model 
%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%
Model.mFile = 'M_solve_RBC';
uqModel = uq_createModel(Model);                % create and add the model object to UQLab

%% histogram
if (false)
    disp('---------------------------------------')
    disp('Calculate histogram of mean and variance of output...')
    %
    X = uq_getSample(100, 'LHS');                  % get parameter sample

    resids = [];                                    % clear residuals
    Y = uq_evalModel(uqModel, X);                   % evaluate model
    save('results/sY_M_V_hist_Results_resids.mat','resids'); % save residuals

    save('results/sY_M_V_hist_Results_X.mat','X');
    save('results/sY_M_V_hist_Results_Y.mat','Y');
    yc = M_solve_RBC(parC);
    M_plot_hist(Y(:,1),outNames{1},'Histogram of Means',       'results/sY_M_hist',yc(1));
    M_plot_hist(Y(:,2),outNames{2},'Histogram of Variances',   'results/sY_V_hist',yc(2));
    m = mean(Y);
    v = var(Y);
    disp(sprintf('Mean    : Mean %2.3f Variance %2.6f Coeff.Var %2.6f',m(1),v(1),sqrt(v(1))/m(1)));
    disp(sprintf('Variance: Mean %2.3f Variance %2.6f Coeff.Var %2.6f',m(2),v(2),sqrt(v(2))/m(2)));
end

%% Sensitivity Analysis
%%%%%%%%%%%%%%%%%%%%%%%%%

%% Sobol MC
if(false)
    disp('---------------------------------------')
    disp('Calculate (MC) Sobol indices of mean and variance of output...')
    %
    clear Opts;
    Opts.Type = 'uq_sensitivity';
    Opts.Method = 'Sobol';
    Opts.Sobol.Order = 1;
    Opts.Sobol.SampleSize = 10;
    Opts.Bootstrap.Replications = 100;
    Results = uq_createAnalysis(Opts); 
    uq_print(Results,1);
    uq_print(Results,2);
    save('results/sY_M_V_Sens_Sobol_MC_Results.mat','Results');
    M_plot_sens(Results.Results.Total,          outNames,parNames,'Total Sobol Indices',        'results/sY_M_V_Sens_Sobol_MC_total',Results.Results.Bootstrap.Total.CI);
    M_plot_sens(Results.Results.AllOrders{1},   outNames,parNames,'First Order Sobol Indices',  'results/sY_M_V_Sens_Sobol_MC_1',Results.Results.Bootstrap.AllOrders{1}.CI);
    %names = M_diname(parNames,Results.Results.VarIdx{2});
    %M_plot_sens(Results.Results.AllOrders{2},   outNames,names,   'Second Order Sobol Indices', 'results/sY_M_V_Sens_Sobol_MC_2',Results.Results.Bootstrap.AllOrders{2}.CI);
end
%% Sobol PCE
disp('---------------------------------------')
disp('Calculate (PCE) Sobol indices of mean and variance of output...')
%
pceOpts.Type = 'uq_metamodel';
pceOpts.MetaType = 'PCE';
pceOpts.FullModel = uqModel;
pceOpts.Degree = 1:20;
pceOpts.Method = 'LARS'; 
pceOpts.ExpDesign.Sampling = 'Sobol';
pceOpts.ExpDesign.NSamples = 500;
pceOpts.Bootstrap.Replications = 100;
pceOpts.LARS.LarsEarlyStop = false;
myPCE = uq_createModel(pceOpts);
clear Opts;
Opts.Type = 'uq_sensitivity';
Opts.Method = 'Sobol';
Opts.Sobol.Order = 2;
Results = uq_createAnalysis(Opts);
uq_print(Results,1);
uq_print(Results,2);
%uq_display(Results);
save('results/sY_M_V_Sens_Sobol_PCE_Results.mat','Results');
M_plot_sens(Results.Results.Total,          outNames,parNames,'Total Sobol Indices',        'results/sY_M_V_Sens_Sobol_PCE_total',Results.Results.Bootstrap.Total.CI);
if (false) 
    M_plot_sens(Results.Results.AllOrders{1},   outNames,parNames,'First Order Sobol Indices',  'results/sY_M_V_Sens_Sobol_PCE_1',Results.Results.Bootstrap.AllOrders{1}.CI);
end
names = M_diname(parNames,Results.Results.VarIdx{2});
M_plot_sens(Results.Results.AllOrders{2},   outNames,names,    'Second Order Sobol Indices','results/sY_M_V_Sens_Sobol_PCE_2',Results.Results.Bootstrap.AllOrders{2}.CI);

%% histogram
disp('---------------------------------------')
disp('Calculate histogram of mean and variance of output using PCE')
%
X = uq_getSample(1000000, 'MC');                  % get parameter sample
Y = uq_evalModel(myPCE, X);                   % evaluate model using PCE
save('results/sY_M_V_hist_Results_X_PCE.mat','X');
save('results/sY_M_V_hist_Results_Y_PCE.mat','Y');
yc = M_solve_RBC(parC);
M_plot_hist(Y(:,1),[],[],       'results/sY_M_hist_PCE',yc(1)); % outNames{1},'Histogram of Means'
M_plot_hist(Y(:,2),[],[],   'results/sY_V_hist_PCE',yc(2)); % outNames{2},'Histogram of Variances'
m = mean(Y);
v = var(Y);
disp(sprintf('Mean    : Mean %2.3f Variance %2.6f Coeff.Var %2.6f',m(1),v(1),sqrt(v(1))/m(1)));
disp(sprintf('Variance: Mean %2.3f Variance %2.6f Coeff.Var %2.6f',m(2),v(2),sqrt(v(2))/m(2)));

%% univariate effects

uq_PCE_displayEE(myPCE) % first order interactions, 1st output is selected (in case of vector output)
saveas(gcf,'results/univariate.fig');
saveas(gcf,'results/univariate.png');
close(gcf);


toc
diary off

uq_saveSession(UQSESSION);