%% Code accompanying "Uncertainty Quantification and Global Sensitivity Analysis for Economic Models"
%% by Daniel Harenberg, Stefano Marelli, Bruno Sudret, and Viktor Winschel

function Y = M_solve_RBC(X,~)
global resids
global additional_QoIs % compute more than just the first two QoIs.

% define parameters of the simulations for the variable of interest
nyrs = 3000; % 3000 5000 10000       % number of periods to simulate
nfirst = 1000; % 1000      % first period from which to calculate expected value 
nrep = 1;           % number of repetitions of simulations
timing = false;

% The following are the empirical counterparts, taken from King, Rebelo (1999), p. 938, Table 1.
varDat = 1.81^2;    % empirical var log production 
                    % (normalizes simulated var log output)
var_rel_cons_Dat = 0.74^2; % In the data, the variance of log consumption relative to log production
var_cons_Dat = 1.35^2; % empirical variance of log consumption

nobs = size(X,1);   % number of parameter values to solve at
Y = zeros(nobs,2);  % return two outputs

wb = waitbar(0,'Please wait...');
for ip=1:nobs;
    waitbar( ip / nobs)

    % get model parameters
    beta =  X(ip,1);   % Discount factor                         
    tau  =  X(ip,2);   % Elasticity of intertemporal substitution
    theta = X(ip,3);   % Consumption and leisure substitution rate
    alpha = X(ip,4);   % Technical substitution rate              
    delta = X(ip,5);   % Depreciation rate                        
    phi   = X(ip,6);   % Capital adjustment cost parameter        
    rho   = X(ip,7);   % Autocorrelation of productivity          
    sigma = X(ip,8);   % Standard deviation of productivity shock 

    % model dimensions
    ds = 2;         % number of states
    dx = 3;         % number of decision variables
    dz = 1;         % number of forward looking variables

    % COMPUTE SHOCK DISTRIBUTION
    m = 3;                                              % discretization
    [e,w] = qnwnorm(m,-(sigma^2)/2,sigma^2);                       % normal nodes and weights

    % PACK MODEL STRUCTURE
    clear model
    model.func = 'M_RBC';                                % model functions
    model.discount = beta;                              % discount factor
    model.e = e;                                        % shocks
    model.w = w;                                        % probabilities
    model.params = {beta tau theta alpha delta phi rho};% other parameters
    model.explicit = 0;                                 % h(s,x,z,sn,xn) not solvable for x(s,z)
    model.noxnext = 0;

    % solve for deterministic steady state
    [kss,ass,css,lss,iss,yss] = M_steady_state(beta,tau,theta,alpha,delta,phi,rho,sigma);

    % linearize and solve
    sss = [kss ass];
    xss = [css lss iss];
    zss = feval(model.func,'h',[],[],[],[],sss,xss,model.params{:});
    ess = 0;

    % prepare matrices A and B
    [f, fs, fx, fz] = feval(model.func,'f',sss,xss,zss,ess,[] ,[] ,model.params{:});
    [g, gx, gs]     = feval(model.func,'g',sss,xss,zss,ess,[] ,[] ,model.params{:});    
    [h, hx, hsn, hxn]     = feval(model.func,'h',[] ,[] ,[] ,ess,sss,xss,model.params{:});    
    fz = squeeze1(fz);
    hsn = squeeze1(hsn);
    hxn = squeeze1(hxn);
    fs = squeeze1(fs);
    fx = squeeze1(fx);
    gs = squeeze1(gs);
    gx = squeeze1(gx);
    A = [ eye(ds,ds), zeros(ds,dx); -fz*hsn , -fz*hxn    ];
    B = [ gs , gx ; fs , fx];
    % qzdecomposition
    [S,T,Q,Z]=qzordered(A,B);
    % policy function
    C=real(Z(ds+1:end,1:ds)/Z(1:ds,1:ds));
    % state transition
    P=real((Z(1:ds,1:ds)*(S(1:ds,1:ds)\T(1:ds,1:ds)))/Z(1:ds,1:ds));   % Z_11 / S_11 * T_11 / Z_11
    %[C1,P1,D,A,B] = relin(model,sss,xss); % same resul as our computation...

    % DEFINE APPROXIMATION SPACE
    n      = [11 11];                                         % degree of approximation
    dmin   = [kss*0.7 min(e)/(1-rho)];                  % state bounds
    dmax   = [kss*1.6 max(e)/(1-rho)];                  % state bounds
    fspace = fundefn('cheb',n,dmin,dmax);           % function space
    snode  = funnode(fspace);                           % state collocaton nodes
    snodes = gridmake(snode);			                % state collocation grid points
    ns = size(snodes,1);
    % Use linearized solution as starting values
    xinit = repmat(xss,ns,1) + (snodes - repmat(sss,ns,1)) * C';

    % SOLVE RATIONAL EXPECTATIONS EQULIBRIUM
    % Set algorithm options  
    options=struct('usebroyden',  0,...
                 'expapprox',   0,...
                 'lowmemory',   0,...
                 'maxit',     40,...
                 'stepsize',  1.5,...
                 'showiters',   0,...
                 'checks',      0,...
                 'nres',        5);

    c=funfitxy(fspace,snodes,xinit);

    if (timing)
        disp('Solving model');
        tic
    end
    [c,rscoord,rx,rh,rf,rresid] = resolve(model,fspace,options,c);
    %resids = [resids rresid];
    if (timing) 
        toc
        disp('Simulating model');
        tic
    end
    % Compute expected state and policy paths of 
    % linear and nonlinear deicisions on the same realisation of shocks 
    sinit = repmat(sss,nrep,1);   
    [spath,xpath] = M_resimul(model,fspace,c,C,sss,xss,sinit,nyrs,true);
    if (timing) 
        toc 
    end
    
    k_path = squeeze(spath(:,1,:));
    %min(k_path) > dmin(1) && max(k_path) < dmax(1) % check sim to be
    %within state space bounds

    a_path = squeeze(spath(:,2,:));
    c_path = squeeze(xpath(:,1,:));
    l_path = squeeze(xpath(:,2,:));
    i_path = squeeze(xpath(:,3,:));

    y_path = (exp(a_path) .* k_path .^ alpha .* l_path .^ (1-alpha));            
        
    normalize = 3.0;
    varDat=0.0013; % because we don't use HP filter.

    % main ones
    %Y(ip,1)=mean(y_path(nfirst:end));                   % old QoI 1
    Y(ip,1)= mean(k_path(nfirst:end)./y_path(nfirst:end))/normalize;  %  mean capital-output ratio    
    Y(ip,2)= var(log(y_path(nfirst:end))) / varDat;      % variance ratio of production (old QoI 2)

    if (additional_QoIs) % compute more than just the first two QoIs.
        Y(ip,3)= var(log(c_path(nfirst:end)))/ var_cons_Dat;      % variance ratio of log consumption (very similar to log production)
        corr = corrcoef(log(y_path(nfirst:end)), log(c_path(nfirst:end)));
        Y(ip,4)= corr(1,2);
        c_growth= c_path(nfirst+1:end)./c_path(nfirst:end-1) - 1;
        Y(ip,5)= var(c_growth); % suggested by R1    
        corr = corrcoef(y_path(nfirst:end), i_path(nfirst:end));
        Y(ip,6)= corr(1,2);
        corr = autocorr(i_path(nfirst:end),1);
        Y(ip,7)= corr(2);
        %Y(ip,3)= var(log(c_path(nfirst:end)))/var(log(y_path(nfirst:end))); % / var_rel_cons_Dat     % relative variance of consumptions        
    end

end
close(wb);


function m = squeeze1(t3d)
s = size(t3d);
m = reshape(t3d,[s(2:end) 1]);
