%% Code accompanying "Uncertainty Quantification and Global Sensitivity Analysis for Economic Models"
%% by Daniel Harenberg, Stefano Marelli, Bruno Sudret, and Viktor Winschel
% Last modified: 29. March 2018

%% SCRIPT CONFIG AND STARTUP
clearvars
% the UQLab framework is needed to run this script. It can be downloaded
% from www.uqlab.com
uqlab

% CONFIGURATION OPTIONS

% Dataset
DESIGN = 'lhs-allQoI';

% Calculate the PCE or load saved ones (very time consuming). If false they
% will be loaded from the session file (UQSESSIONFILE)
CALCULATE = true;
% Process the PCEs to create convergence plots and sensitivity analyses 
% If set to false they will be loaded from the PROCESSEDFILENAME file
PROCESS = true;
PROCESSEDFILENAME = sprintf('ConvergenceAnalysis_AllQoIs_%s.mat',DESIGN);
UQSESSIONFILE = 'UQLab_FullSession_allQoI.mat';
UQINPUTMODELFILE = 'UQLab_FullModel_Input_and_Model_allQoI.mat';

% Create an experimental design
CREATEED = true;
EDFILE = 'UQLab_ED_EnrichLHS_50-10-500_allQoI.mat';
if CREATEED
    uq_FullModel_CreateED
else
    load(EDFILE);
end
M = size(XED,2);
% Plot the results into a large set of figures
PLOT = true;

% Select which outputs should be plotted
PLOTOUTS = 1:6;
NOUT = length(PLOTOUTS);

% Plot elementary plots for all QoIs in the same plot
SAMEPLOT = true;
% Portrait or landscape QoI plot
LANDSCAPE = true;
% Generate a single figure per output
SEPARATEFIGS = false;

% COLORMAPs
COLS = [0 0 1; 1 0 0; uq_cmap(5)];
COLS = COLS(PLOTOUTS,:);


% Experimental design sizes used for convergence study
EDSIZES = 50:10:500;
NEDS = length(EDSIZES);

% Plot histograms of QoIs
PLOTHIST = false;
NHIST = 1e6; % Number of points to generate the histograms

% Labels for plots (output names)
LABELS = {'Average capital-output ratio', ...
                'Production variance ratio', ...
                'Consumption variance ratio', ...
                'Corr(cons,production)', ...
                'Variance consumption growth',... %}; 
                'Corr(investment,production)', ...
                'Autocorr(investment)'};
LABELS = LABELS(PLOTOUTS);


% Labelling for the inputs (used, e.g. for Sobol' indices)
VARNAMES = {'\beta','\tau', '\chi','\alpha','\delta', '\phi', '\rho','\sigma'};
VARIDX2ND = [];
VARNAMES2ND = {};
for ii = 1:7
    for jj = ii+1:8
        VARIDX2ND = [VARIDX2ND; [ii jj]];
        VARNAMES2ND = [VARNAMES2ND; [VARNAMES{ii} VARNAMES{jj}]];
    end
end


%% Startup UQLab and load the input (only if calculating or processing)
if CALCULATE || PROCESS
    uqlab(UQINPUTMODELFILE);
end


%% CALCULATION OF THE PCE MODELS
if CALCULATE
    % Load the input and the model
    myInput = uq_getInput;
    FullModel = uq_getModel;
    
    % Number of inputs
    M = length(myInput.Marginals);
    
    % Load the ED (creates XED and YED)
    load(EDFILE);
    NPCOUT = size(YED,2);
    
     
    %% Create the metamodels and the sensitivities
    % common PCE options
    PCEOpts.Type = 'Metamodel';
    PCEOpts.MetaType = 'PCE';
    PCEOpts.Input = myInput;
    PCEOpts.LARS.LarsEarlyStop = false;
    PCEOpts.Degree = 1:20;
    PCEOpts.Bootstrap.Replications = 100; % Enable bootstrap
    
    % common sensitivity options
    SensOpts.Type = 'Sensitivity';
    SensOpts.Method = 'Sobol';
    SensOpts.Sobol.Order = 2;


    % Loop over the EDs
    wb = waitbar(0, 'Running...');
    for ii = 0:(NEDS-1)
        % Subset of full design
        idx = 1:EDSIZES(ii+1);
        % Specify the name
        PCEOpts.Name = sprintf('PCE %s %d',DESIGN,idx(end));
        % Specify the ED
        PCEOpts.ExpDesign.X = XED(idx,:);
        PCEOpts.ExpDesign.Y = YED(idx,:);
        % Create the PCE
        myPCE{ii+1} = uq_createModel(PCEOpts,'-overwrite');
        % Create the Sensitivity
        SensOpts.Name = sprintf('Sobol Indices %s %d',DESIGN,idx(end));
        mySens{ii+1} = uq_createAnalysis(SensOpts);
        waitbar(ii/45, wb, sprintf('Running: ED %d', idx(end)));
    end % ii: loop over designs
    close(wb)
    
    % Save the PCEs for later re-use
    uq_saveSession(UQSESSIONFILE);
end

%% CONVERGENCE ANALYSIS
% Only load the uqlab session file if the whole thing has not been already
% processed
if PROCESS
    % Load the UQLab session (containing the PCEs) if they were not
    % calculated
   if ~CALCULATE
       uqlab(UQSESSIONFILE);
   end
   
   % Initialize the containers
   LOO = zeros(NEDS,NOUT); % LOO error
   
   TotalSobol = zeros(NEDS,M,NOUT);  % Total Sobol indices
   TotalSobolHi = TotalSobol;   % Upper bounds
   TotalSobolLo = TotalSobol;   % Lower bounds
   
   FirstOrderSobol = zeros(NEDS,M,NOUT);  % First Order Sobol indices
   FirstOrderSobolHi = FirstOrderSobol;   % Upper bounds
   FirstOrderSobolLo = FirstOrderSobol;   % Lower bounds
   
   SecondOrderSobol = zeros(NEDS,nchoosek(M,2),NOUT);  % Second Order Sobol indices
   SecondOrderSobolHi = SecondOrderSobol;   % Upper bounds
   SecondOrderSobolLo = SecondOrderSobol;   % Lower bounds
   
   % Loop over the experimental designs
   for ii = 0:(NEDS-1)
       % Get the model and the sensitivity based on the stored filenames
       myPCE = uq_getModel(sprintf('PCE %s %d',DESIGN,EDSIZES(ii+1)));
       mySens = uq_getAnalysis(sprintf('Sobol Indices %s %d',DESIGN,EDSIZES(ii+1)));
       
       % Getting some diagnostic info
       for oo = 1:NOUT
           oidx = PLOTOUTS(oo);
           LOO(ii+1,oo) = myPCE.Error(oidx).LOO;
       end
       
       
       % Getting the Sobol' indices
       TotalSobol(ii+1,:,:) = mySens.Results(end).Total(:,PLOTOUTS);
       TotalSobolLo(ii+1,:,:) = mySens.Results(end).Bootstrap.Total.CI(:,PLOTOUTS,1);
       TotalSobolHi(ii+1,:,:) = mySens.Results(end).Bootstrap.Total.CI(:,PLOTOUTS,2);
       
       FirstOrderSobol(ii+1,:,:) = mySens.Results(end).FirstOrder(:,PLOTOUTS);
       FirstOrderSobolLo(ii+1,:,:) = mySens.Results(end).Bootstrap.FirstOrder.CI(:,PLOTOUTS,1);
       FirstOrderSobolHi(ii+1,:,:) = mySens.Results(end).Bootstrap.FirstOrder.CI(:,PLOTOUTS,2);
       
       SecondOrderSobol(ii+1,:,:) = mySens.Results(end).AllOrders{2}(:,PLOTOUTS);
       SecondOrderSobolLo(ii+1,:,:) = mySens.Results(end).Bootstrap.AllOrders{2}.CI(:,PLOTOUTS,1);
       SecondOrderSobolHi(ii+1,:,:) = mySens.Results(end).Bootstrap.AllOrders{2}.CI(:,PLOTOUTS,2);
       
       
   end % ii: loop over EDs
   
   % save the processed results
   save(PROCESSEDFILENAME, 'TotalSobol*', '*OrderSobol*','LOO', '*ED*');
else % if PROCESS
    % load the results if already processed
    load(PROCESSEDFILENAME);
end

%% PLOT THE RESULTS
if PLOT
    if ~CALCULATE && ~PROCESS
        uqlab(UQSESSIONFILE);
    end
    NIDX = 2;
    NSIDX = 2;
    %% First the convergence of the three largest Total indices
    clear tidx;
    for oo = 1:NOUT
        [~, tidx(:,oo)] = sort(TotalSobol(end,:,PLOTOUTS(oo)), 'descend');
    end
    tidx = tidx(1:NIDX,:);
    
    % Plot them against the ED size
    FIGNAME = 'TotalConvergence';
    uq_figure('Name', FIGNAME, 'filename',FIGNAME);  
    hold on;
    for ii = 1:NIDX
        for oo = 1:NOUT % loop over the output
            SI = TotalSobol(:,tidx(ii,oo),PLOTOUTS(oo));
            SIL = TotalSobolLo(:,tidx(ii,oo),PLOTOUTS(oo));
            SIH = TotalSobolHi(:,tidx(ii,oo),PLOTOUTS(oo));
            EB(oo) = errorbar(EDSIZES', SI, SI-SIL, SIH-SI, 'color',COLS(oo,:));
            if mod(oo,2)
                xx = EDSIZES(end)*.5;
            else
                xx = EDSIZES(end)*.55;
            end
            yy = SIH(end)+.04;
            
            t = text(xx,yy,sprintf('%s',VARNAMES{tidx(ii,oo)}), 'Color',COLS(oo,:),'FontSize',16, 'FontWeight','bold');
        end
    end
    box on;
    ylim([0 1.]);
    xlim([40 510])
    legend(EB, LABELS, 'location', 'northeast')
    xlabel('N')
    ylabel('S^T_i')
    
    %% Convergence of the three largest first order indices
    clear tidx
    for oo = 1:NOUT
        [~, tidx(:,oo)] = sort(FirstOrderSobol(end,:,PLOTOUTS(oo)), 'descend');
    end
    
    tidx = tidx(1:NIDX,:);
    
    % Plot them against the ED size
    FIGNAME = 'FirstOrderConvergence';
    uq_figure('Name', FIGNAME, 'filename',FIGNAME);  
    hold on;
    for ii = 1:NIDX % loop over the indices
        for oo = 1:NOUT % loop over the output
            SI = FirstOrderSobol(:,tidx(ii,oo),PLOTOUTS(oo));
            SIL = FirstOrderSobolLo(:,tidx(ii,oo),PLOTOUTS(oo));
            SIH = FirstOrderSobolHi(:,tidx(ii,oo),PLOTOUTS(oo));
            EB(oo) = errorbar(EDSIZES', SI, SI-SIL, SIH-SI, 'color',COLS(oo,:));
            if mod(oo,2)
                xx = EDSIZES(end)*.5;
            else
                xx = EDSIZES(end)*.55;
            end
            yy = SIH(end)+.045;
            t = text(xx,yy,sprintf('%s',VARNAMES{tidx(ii,oo)}), 'Color',COLS(oo,:),'FontSize',16, 'FontWeight','bold');
            
                        
        end
    end
    box on
    
    ylim([0 1.]);
    xlim([40 510])
    legend(EB, LABELS, 'location', 'northeast')
    xlabel('N')
    ylabel('S_i')
    
    %% Convergence of two second order indices 
    clear tidx
    for oo = 1:2
        [~, tidx(:,oo)] = sort(SecondOrderSobol(end,:,PLOTOUTS(oo)), 'descend');
    end
    tidx = tidx(1:NIDX,:);
    % select manually the specific indices
    ttidx = [1 1; 3 2];
    
    % Plot them against the ED size
    FIGNAME = 'SecondOrderConvergence';
    uq_figure('Name', FIGNAME, 'filename',FIGNAME);  
    hold on;
    VARNAMES2D = {'\beta\delta', '\rho\sigma'; '\beta\alpha','\alpha\rho'};
    for ii = 1:NSIDX
        for oo = 1:2 % loop over the output
            SIDX = EDSIZES >= 120;
            SI = SecondOrderSobol(SIDX,tidx(ii,oo),PLOTOUTS(oo));
            SIL = SecondOrderSobolLo(SIDX,tidx(ii,oo),PLOTOUTS(oo));
            SIH = SecondOrderSobolHi(SIDX,tidx(ii,oo),PLOTOUTS(oo));
            EB(oo) = errorbar(EDSIZES(SIDX)', SI, SI-SIL, SIH-SI,'color',COLS(oo,:));
            xx = EDSIZES(1) + 0.55*range(EDSIZES);
            yy = SIH(ceil(end/2))+.003;

            t = text(xx,yy,sprintf('%s',VARNAMES2D{ii,oo}), 'Color',COLS(oo,:),'FontSize',16, 'FontWeight','bold');

        end
    end
    
    
    
    %text(0.5)
    box on;
    ylim([0 0.08]);
    xlim([100 510])
    legend(EB, LABELS, 'location', 'northeast')
    xlabel('N')
    ylabel('S_{ij}')
    
    %% Plot the convergence of the LOO error
    FIGNAME = 'LOO_Convergence';
    uq_figure('name', FIGNAME, 'Filename', FIGNAME)
    hold on;
    for oo = 1:NOUT
        P_loo(oo) = plot(EDSIZES', LOO(:,oo), 'color',COLS(oo,:));
    end
    set(gca,'yscale', 'log');
    box on
    xlim([40 510])
    legend(P_loo, LABELS, 'location', 'northeast')
    xlabel('N')
    ylabel('\it err_{LOO}')
end

%% HISTOGRAMS
if PLOTHIST
    % Get the PCE with 150 samples in the ED (number 12)
    myPCE = uq_getModel(12);
    % get the model evaluations on a set of 1e6 samples
    X = uq_getSample(1e6,'Sobol');
    Y = uq_evalModel(myPCE,X);
    % go to the plotting
    
    for oo = 1:NOUT
        FIGNAME = sprintf('Y_Histogram_%d',PLOTOUTS(oo));
        uq_figure('name',FIGNAME,'filename',FIGNAME);
        hold on;
        tt = uq_histogram(Y(:,PLOTOUTS(oo)));
        axis tight
        xlabel(LABELS{oo});
    end
end

%% Total Sobol' indices (150 Samples)


if ~SAMEPLOT
    for oo = 1:NOUT
        FIGNAME = sprintf('TFSobolIndices_%d',PLOTOUTS(oo));
        uq_figure('Name', FIGNAME, 'FileName', FIGNAME)
        TS = squeeze(TotalSobol(12,:,PLOTOUTS(oo)));
        TSU = squeeze(TotalSobolHi(12,:,PLOTOUTS(oo)));
        TSL = squeeze(TotalSobolLo(12,:,PLOTOUTS(oo)));
        FS = squeeze(FirstOrderSobol(12,:,PLOTOUTS(oo)));
        FSU = squeeze(FirstOrderSobolHi(12,:,PLOTOUTS(oo)));
        FSL = squeeze(FirstOrderSobolLo(12,:,PLOTOUTS(oo)));
        bb(2) = uq_bar(TS,'edgecolor', 'none', 'facecolor', 'r');
        hold on;
        bb(1) = uq_bar(FS,'edgecolor', 'none', 'facecolor', 'b');
        set(gca,'XTickLabel',VARNAMES, 'XTick',1:8,'fontsize',18,'box', 'on')
        xlim([0 9]);
        ylim([0 1]);
        legend(bb,'S_i', 'S^T_i')
    end
else % Plot all the indices in the same plot, side by side
    FIGNAME = sprintf('TFSobolIndices_Combined_%d',PLOTOUTS(oo));
    uq_figure('Name', FIGNAME, 'FileName', FIGNAME)
    axes;hold on;
    dx = 1/NOUT;
    M = size(TotalSobol,2);
    SQ = 1.5;
    xx = (1:M)*SQ;
    for oo = 1:NOUT
        TS = squeeze(TotalSobol(12,:,PLOTOUTS(oo)));
        TSU = squeeze(TotalSobolHi(12,:,PLOTOUTS(oo)));
        TSL = squeeze(TotalSobolLo(12,:,PLOTOUTS(oo)));
        bb(oo) = bar(xx+dx*(oo-1),TS,dx/SQ,'edgecolor', 'none', 'facecolor', COLS(oo,:));
        %         hold on;
        %         bb(1) = uq_bar(FS,'edgecolor', 'none', 'facecolor', 'b');
    end
    set(gca,'XTickLabel',VARNAMES, 'XTick',xx+dx,'fontsize',18,'box', 'on')
    xlim([min(xx)-1 max(xx)+1]);
    ylim([0 1]);
    legend(bb,LABELS,'fontsize',14,'location','northwest')
end


%% 2nd order Sobol' indices (250 Samples)

for oo = 1:NOUT
    FIGNAME = sprintf('SecondOrderSobolIndices_%d', PLOTOUTS(oo));
    uq_figure('Name', FIGNAME, 'FileName', FIGNAME)
    SS = squeeze(SecondOrderSobol(21,:,PLOTOUTS(oo)));
    [SS, SSIDX] = sort(SS, 'descend');
    SSU = squeeze(SecondOrderSobolHi(21,SSIDX,PLOTOUTS(oo)));
    SSL = squeeze(SecondOrderSobolLo(21,SSIDX,PLOTOUTS(oo)));
    uq_bar(SS(1:NSIDX), 'edgecolor', 'none', 'facecolor', COLS(oo,:))
    hold on;
    
    set(gca,'XTickLabel',VARNAMES2ND(SSIDX(1:NSIDX)), 'XTick',1:NSIDX)
    xlim([0 NSIDX+1]);
    ylim([0 0.06]);
end

%% Univariate effects (150 samples)
myPCE = uq_getModel(12);
X = uq_getSample(1e3, 'Sobol');
if LANDSCAPE
    W = 1200;
    H = 600;
else
    W = 600;
    H = 800;
end
for oo = 1:NOUT
    myPCE_EE = uq_PCE_createEEPCE(myPCE,1,PLOTOUTS(oo));
    YEE = uq_evalModel(myPCE_EE,X);
    
    yLimits = [min(YEE(:)) max(YEE(:))];
    FIGNAME = sprintf('UnivariateEffects_%d_sameplot_%i_landscape_%i_NOUT_%i',PLOTOUTS(oo),SAMEPLOT,LANDSCAPE,NOUT);
    
    if ~SEPARATEFIGS
        if ~SAMEPLOT
            uq_figure('filename', FIGNAME, 'name',FIGNAME, 'position',[(oo-1)*W 100 W H]);
        else
            if oo == 1
                uq_figure('filename', FIGNAME, 'name',FIGNAME, 'position',[(oo-1)*W 100 W H]);
            end
        end
    end
    
    for ii = 1:8 % LOOP OVER THE INPUTS
        if SAMEPLOT
            % normalize the Elementary effects if on the same plot
            YEE(:,ii) = (YEE(:,ii) - myPCE.PCE(PLOTOUTS(oo)).Moments.Mean)/sqrt(myPCE.PCE(PLOTOUTS(oo)).Moments.Var);
            if oo == 1
                if ~SEPARATEFIGS
                    if LANDSCAPE
                        sp(ii) = subplot(2,4,ii);
                    else
                        sp(ii) = subplot(4,2,ii);
                    end
                else % create a separate plot for each variable
                    FIGNAMESEP = sprintf('%s_separate_X%i',FIGNAME,ii);
                    uq_figure('filename', FIGNAMESEP, 'name',FIGNAMESEP);
                    sp(ii) = axes;
                end
            
            else
                axes(sp(ii));
            end
        else % SAMEPLOT
            if LANDSCAPE
                sp(ii) = subplot(2,4,ii);
            else
                sp(ii) = subplot(4,2,ii);
            end % LANDSCAPE
        end
        
        
        if ~SEPARATEFIGS && oo == 1
            P = get(gca, 'Position');
            P(3) = P(3)*0.95;
            P(4) = P(4)*0.95;
            set(gca, 'Position', P);
        end
        set(gca,'NextPlot','add');
        
        uq_plot(X(:,ii), YEE(:,ii), '.', 'color', COLS(oo,:));
        set(gca, 'fontsize', 12);
        xlabel(VARNAMES{ii}, 'fontsize', 20);
        ylabel(sprintf('$M_{%s}$',VARNAMES{ii}), 'fontsize', 20, 'interpreter', 'latex');
        axis tight

        YL = ylim;
        if SAMEPLOT
            ylim([YL(1) - 0.1*range(YL), YL(2) + 0.1*range(YL)]);
        else
            ylim(yLimits);
        end
    end % LOOP OVER THE INPUTS
    
end % LOOP OVER THE OUTPUTS


if ~SEPARATEFIGS && LANDSCAPE 
    ghostplots = [];
    for oo = 1:NOUT
        hold on;
        ghostplots = [ghostplots plot(nan,nan,'-','color', COLS(oo,:))];
    end
    legend(ghostplots,LABELS,'orientation', 'horizontal', 'position', [0.13 0.93 0.768 0.06])
end